﻿define([
"dojo/_base/array",
"dojo/_base/declare",
"dojo/_base/Deferred",
"dojo/_base/lang",
"dojo/store/Memory",
"dojo/topic",
"dojo/when",
"epi/shell/_ContextMixin",
"epi/shell/layout/AnimatedCardContainer",
"epi/shell/TypeDescriptorManager"
], function (
array,
declare,
Deferred,
lang,
Memory,
topic,
when,
_ContextMixin,
AnimatedCardContainer,
TypeDescriptorManager
) {

    // module:
    //		epi/shell/widget/WidgetSwitcher
    // summary:
    //		A container that handles showing a child widget based on either a
    //		specific view request or a change in the current context.

    return declare([AnimatedCardContainer, _ContextMixin], {
        // summary:
        //		A container that handles showing a child widget based on either a
        //		specific view request or a change in the current context.

        // componentConstructorArgs: Object
        //      Arguments that must be passed to constructor when new component is created.
        componentConstructorArgs: null,

        // supportedContextTypes: Array
        //      String array of supported context data types.
        //      This option should be defined if we need to react only on context changes with specific data type.
        //      Context changes are ignored if set to empty array.
        supportedContextTypes: null,

        constructor: function (args) {
            // summary:
            //      Sets up the store.

            this.store = this.createStore();
            this.own(this.store);
        },

        postMixInProperties: function () {
            // summary:
            //      Verifies and fixes the list of supported context data types.
            // tags:
            //      protected
            this.inherited(arguments);
            if (this.supportedContextTypes) {
                if (!lang.isArray(this.supportedContextTypes)) {
                    this.supportedContextTypes = [this.supportedContextTypes];
                }
                if (!array.every(this.supportedContextTypes, function (type) { return typeof type == "string"; })) {
                    throw new Error("supportedContextTypes must be string, array of strings or null.");
                }
            }
        },

        postCreate: function () {
            // summary:
            //		If the current context is available it loads the initial view component for it. Also subscribes
            //		to view component change requests.
            // tags:
            //		protected

            this.inherited(arguments);

            Deferred.when(this.getCurrentContext(), lang.hitch(this, this.contextChanged));

            this.own(topic.subscribe("/epi/shell/action/changeview", lang.hitch(this, "viewComponentChangeRequested")));
            this.own(topic.subscribe("/epi/shell/action/changeview/back", lang.hitch(this, "_viewComponentBackRequested")));
        },

        contextChanged: function (/*Object*/context, data) {
            // summary:
            //		Called when the current context changes.
            // tags:
            //		protected

            if (!this._isContextTypeSupported(context)) {
                return;
            }

            var view = context.customViewType || (!!data ? data.viewType : null);
            var currentWidgetInfo = this._getCurrentWidgetInfo();
            //NOTE: when we change so that we can route to different views this should be rewritten.
            if (data && data.contextIdSyncChange) {
                currentWidgetInfo && (view = currentWidgetInfo.type);
            }

            if (!view && context.dataType) {
                // If there is no custom view type, fall back to the widget for the type.
                // For backward compatibility, check the obseleted property mainWidgetType
                view = TypeDescriptorManager.getValue(context.dataType, "mainWidgetType");
            }

            if (view) {
                this._getObject(view, null, context, data);
            } else {
                var suggestedView = TypeDescriptorManager.getValue(context.dataType, "defaultView"),
                    availableViews = this._getAvailableViews(context.dataType),
                    requestedViewName = data ? data.viewName : null;

                // TODO: consider removing reference to "view"(preview) when refactoring the view rerouting
                if (currentWidgetInfo && currentWidgetInfo.viewName == "view"){
                    // if we're viewing the preview we should keep that view as long as it's availiable.
                    array.some(availableViews, function(availableView){
                        if (availableView.key == currentWidgetInfo.viewName){
                            suggestedView = "view";
                            return true;
                        }
                        return false;
                    });
                }

                var viewToLoad = this._getViewByKey(requestedViewName || suggestedView, availableViews);

                if (!viewToLoad) {
                    throw "No default view found for " + context.dataType;
                }

                this._loadViewComponentByConfiguration(viewToLoad, availableViews, null, context, data);
            }
        },

        _getViewByKey: function (key, availableViews) {
            return array.filter(availableViews, function (availableView) {
                return availableView.key === key;
            })[0];
        },

        _getAvailableViews: function (dataType) {
            var availableViews = TypeDescriptorManager.getAndConcatenateValues(dataType, "availableViews"),
                disabledViews = TypeDescriptorManager.getValue(dataType, "disabledViews");

            var filteredViews = array.filter(availableViews, function (availableView) {
                return array.every(disabledViews, function (disabledView) {
                    return availableView.key !== disabledView;
                });
            });
            return filteredViews.sort(function (x, y) {
                return x.sortOrder < y.sortOrder ? -1 : 1;
            });
        },

        _loadViewComponentByConfiguration: function (view, availableViews, args, context, data) {
            data = lang.mixin(data, {
                viewName: view.key,
                availableViews: availableViews
            });

            this._getObject(view.controllerType, args, context, data);
        },

        viewComponentChangeRequested: function (/*String*/type, /*Object*/args, /*Object*/data) {
            // summary:
            //		Called when a new view component is requested.
            // type: String
            //      Specify component widget type or a key which refers to a view configuration.
            // args: Object
            //      Arguments used to instantiate component widget.
            // data: Object
            //      data passed to the view when calling updateView.
            // tags:
            //		protected

            when(this.getCurrentContext(), lang.hitch(this, function (currentContext) {

                var dataType = currentContext ? currentContext.dataType : null;

                var availableViews = this._getAvailableViews(dataType),
                    view = this._getViewByKey(type, availableViews);

                if (view) {
                    // A view configuration found, means that new view is being requested by key
                    this._loadViewComponentByConfiguration(view, availableViews, args, currentContext, data);
                } else {
                    // Otherwise, by component type. (Backward compatibility)
                    this._getObject(type, args, currentContext, data);
                }
            }));
        },

        _viewComponentBackRequested: function (forceReload) {
            var previousWidgetInfo = this._getPreviousWidgetInfo();
            if (previousWidgetInfo) {
                if (typeof forceReload === 'undefined') {
                    // forceReload should be false by default, which means that we will skip updating view.
                    forceReload = false;
                }
                var data = {
                    skipUpdateView: !forceReload,
                    availableViews: previousWidgetInfo.availableViews,
                    viewName: previousWidgetInfo.viewName
                };
                when(this.getCurrentContext(), lang.hitch(this, function (currentContext) {
                    this._getObject(previousWidgetInfo.type, null, currentContext, data);
                }));
            }
        },

        _getPreviousWidgetInfo: function () {
            // Get previous from sorted descending array.
            var history = this._getWidgetHistory();
            return (history && history.length > 1) ? history[1] : null;
        },

        _getCurrentWidgetInfo: function () {
            var history = this._getWidgetHistory();
            return (history && !!history.length) ? history[0] : null;
        },

        _getWidgetHistory: function () {
            // summary:
            //      Return previous widget.
            // tags:
            //      private

            return this.store.query({}, {
                count: 2, // Get current and previous
                sort: [{
                    attribute: "dateAdded",
                    descending: true
                }]
            });
        },

        _isContextTypeSupported: function (context) {
            // summary:
            //		Verifies if data type of specified context is supported by this instance
            // tags:
            //		private
            if (!context) {
                return false;
            }
            if (!this.supportedContextTypes) {
                return true;
            }
            return array.some(this.supportedContextTypes, function (type) { return context.type === type; });
        },

        _getObject: function (type, args, context, data) {
            // summary:
            //		Gets an instance for the given type if it is different from
            //		the current view component type.
            // tags:
            //		private

            if (!type) {
                return;
            }

            var current = this.selectedChildWidget;

            var func = lang.hitch(this, function () {
                Deferred.when(this._testComponentInstanceOf(current, type), lang.hitch(this, function (result) {
                    if (result) {
                        this._updateStore(type, data, context);
                        this._updateView(current, context, data);
                        this._onViewChanged(type, args, data);
                    }
                    else {
                        this._changeViewComponent(type, args, context, data);
                    }
                }));
            });

            if (current && current.savePendingChanges) {
                Deferred.when(current.savePendingChanges(), func);
            } else {
                func();
            }

        },

        _changeViewComponent: function (type, args, context, data) {
            // summary:
            //		Remove the current view component and add the new one.
            // tags:
            //		private

            Deferred.when(this._getChildByType(type), lang.hitch(this, function (component) {
                var current = this.selectedChildWidget;

                // suspend the current one before changing to the next
                if (current) {
                    current.set("isActive", false);
                }

                if (component) {
                    // activate the component
                    component.set("isActive", true);

                    if (data && data.treatAsSecondaryView) {
                        this.selectSecondaryChild(component);
                    } else {
                        this.selectChild(component);
                    }
                    this._updateStore(type, data, context);
                    this._updateView(component, context, data, true);
                    this._onViewChanged(type, args, data);
                }
                else {
                    require([type], lang.hitch(this, function (componentClass) {
                        var mixedArgs = lang.mixin(lang.mixin({}, this.componentConstructorArgs), args);
                        component = new componentClass(mixedArgs);

                        //component.transitionMode = "reveal";
                        component.fitContainer = true;

                        this.addChild(component);

                        var select;

                        if (data && data.treatAsSecondaryView) {
                            select = this.selectSecondaryChild(component);
                        } else {
                            select = this.selectChild(component);
                        }

                        when(select, lang.hitch(this, function () {
                            this._updateStore(type, data, context);
                            this._updateView(component, context, data, true);
                            this._onViewChanged(type, args, data);
                        }));
                    }));
                }
            }));
        },

        _updateView: function (component, context, data, widgetResumed) {
            if (component && !!!component._started && typeof component.startup === "function") {
                // Ensure component ready.
                component.startup();
            }

            if (component && typeof component.updateView === "function") {
                component.updateView(data, context, { widgetResumed: widgetResumed });
            }
        },

        _getChildByType: function (type) {
            var def = new Deferred();

            require([type], lang.hitch(this, function (componentClass) {
                var components = array.filter(this.getChildren(), function (child, index) {
                    return child.constructor === componentClass;
                }, this);

                def.resolve(components.length ? components[0] : null);
            }));

            return def;
        },

        _testComponentInstanceOf: function (component, type) {
            var def = new Deferred();

            if (component) {
                if (typeof type === "string") {
                    require([type], function (componentClass) {
                        def.resolve(component && component.constructor === componentClass);
                    });
                } else {
                    def.resolve(component && component.constructor === type.constructor);
                }
            } else {
                def.resolve(false);
            }

            return def;
        },

        _onViewChanged: function (type, args, data) {
            topic.publish("/epi/shell/action/viewchanged", type, args, data);
        },

        createStore: function () {
            // summary:
            //      Create new store
            // tags:
            //      protected

            return new Memory({
                data: [],
                idProperty: "id"
            });
        },

        _updateStore: function (/*string*/type, data, context) {
            // summary:
            //      Add data to widget history store
            //      Keep total items in store always 2
            // tags:
            //      private

            var store = this.store;
            store.put(this.createStoreItem(type, data, context), { overwrite: true });

            // If the store contains more than 2 items, remove the oldest one
            if (store.data.length > 2) {
                // get the oldest item
                var results = store.query({}, {
                    sort: [{
                        attribute: "dateAdded",
                        descending: false
                    }]
                });
                store.remove(store.getIdentity(results[0]));
            }
        },

        createStoreItem: function (/*string*/type, data, context) {
            // summary:
            //      Create new object that will be put to store
            // tags:
            //      protected

            var storeItem = { type: type, dateAdded: new Date(), id: type };
            if (data) {
                var availableViews = data.availableViews || (context && this._getAvailableViews(context.dataType));
                var viewName = data.viewName || this._getViewName(availableViews, context);

                storeItem = lang.mixin(storeItem, {
                    availableViews: availableViews,
                    viewName: viewName
                });
                storeItem.id += "#" + viewName;
            }
            return storeItem;
        },

        _getViewName: function(availableViews, context) {
            if (!context) {
                return null;
            }

            var selectedView = this._getViewByKey(TypeDescriptorManager.getValue(context.dataType, "defaultView"), availableViews);
            if (!selectedView) {
                return null;
            }
            return selectedView.key;
        }
    });
});
